
#@author: James V. Talwar

from ..GenomicObjects import Pvar, SSF
import os
import pandas as pd
import logging
import time
from .utils import _QueryGrievousInstall

logger = logging.getLogger(__name__)

'''
About: Helper method to create a new grievous database/dictionary directory.

Input(s):  1) newDictPath: String corresponding to the directory where empty database/dictionary records will be created to be updated.
           2) dictStorage: String corresponding to database storage method. Default: 'parquet'
Output(s): None
'''
def _Create_Dictionary(newDictPath, dictStorage = "parquet"):
    if os.path.isdir(newDictPath): #chromosome-level paralleleization - if database is in process of (or has finished) being created don't need to re-create/overwrite db
        return None

    #Create new dictionary folder
    os.makedirs(newDictPath, exist_ok = True)
    os.makedirs(os.path.join(newDictPath, "Records"), exist_ok = True)

    #Create empty record files and empty dictionary files 
    chromosomes = [str(i) for i in range(1,23)] + ["X", "Y", "MT"] #Grievous standard for chromosomes

    for chromosome in chromosomes:
        # all the files aligned to given database 
        with open(os.path.join(newDictPath, f"Records/CHR_{chromosome}.aligned.record.txt"), "a") as alignedRecord: 
            pass

        # all the files which were used to update the given database. If --shutoff_dict_update is never used then is equivalent to aligned
        with open(os.path.join(newDictPath, f"Records/CHR_{chromosome}.updated.record.txt"), "a") as updatedRecord: 
            pass

        if dictStorage == "parquet":
            parquetWrite = os.path.join(newDictPath, f"CHR_{chromosome}.parquet")
            if not os.path.exists(parquetWrite):
                pd.DataFrame([], columns = ["CHR", "POS", "REF", "ALT", "ID"]).to_parquet(parquetWrite)
            
        else:
            tsvWrite = os.path.join(newDictPath, f"CHR_{chromosome}.tsv")
            if not os.path.exists(tsvWrite):
                pd.DataFrame([], columns = ["CHR", "POS", "REF", "ALT", "ID"]).to_csv(tsvWrite, sep = "\t")
            

    return None

'''
Input(s): dbAlias: String corresponding to an already created or novel dictionary/database alias. 
Outputs: grievousDictDir: String corresponding to the full path of the desired grievous dictionary/database directory.
'''
def _Check_Dictionary(dbAlias):
    grievousDir = _QueryGrievousInstall()

    #check among the dictionary aliases
    grievousDictAliases = [el for el in os.listdir(grievousDir) if os.path.isdir(os.path.join(grievousDir, el))]

    if dbAlias not in grievousDictAliases:
        if dbAlias is None: #Handle automatically when only one dictionary standard exists.
            assert len(grievousDictAliases) == 1, """--database not provided and either more than one database exists or none have been created. 
            Provide a current database/dictionary alias or provide a new one to create a new empty database. Database aliases can be found with grievous list_dbs."""

            grievousDictDir = os.path.join(grievousDir, grievousDictAliases[0])

        else:
            logger.warning(f"CREATING A NEW GRIEVOUS DICTIONARY WITH ALIAS: {dbAlias}")
            
            grievousDictDir = os.path.join(grievousDir, dbAlias)
            _Create_Dictionary(grievousDictDir)

    else:         
        grievousDictDir = os.path.join(grievousDir, dbAlias)

    # Ensure all chromosomal dictionaries/dbs are written; instances of creation and file non-existence can occur when initializing a DB and running in parallel. 
    # To prevent a given chromosome from attempting to access a db before creation (while in progress) sleep to allow for all chromosome creation.
    if not (set([el.split(".")[0].split("_")[1] for el in os.listdir(grievousDictDir) if "CHR" in el]) == set([str(i) for i in range(1,23)] + ["X", "Y", "MT"])):
        time.sleep(13)


    return grievousDictDir


'''
About: This method performs grievous realignment, by creating a PVAR/SSF object and then calling the instance methods. For full details of these methods see
package GenomicObjects.

Input(s): 1) file: String corresponding to the path of the file which the user is attempting to realign.
          2) dictionary: String corresponding to the dictionary/database to which the user is attempting to realign. 
          3) write_path: String corresponding to the directory where the user wants to write realignment files.
          4) mapping: String corresponding to the file path registering all mandatory non-grievous standard column names to the grievous standard.
          5) file_type: String in {'pvar', 'ssf'} corresponding to the desired genomic file type to be realigned.
          6) return_all: Boolean corresponding to whether to return all columns of the underlying file, or the grievous subset.
          7) comment_characters: List of all comment characters corresponding to lines to skip at the beginning of a genomic file.
          8) shutoff_dict_update: Boolean corresponding to whether to skip the dictionary update step. If there is ever any ambiguity 
          as to whether to use this or not (or whether one will align future cohorts), then this parameter SHOULD ALWAYS BE FALSE!

Output(s): None
Write(s): Grievous PVAR/SSF output files (for full details see Pvar.Write() or SSF.Write()).
'''
def Realign(file, dictionary, write_path, mapping, file_type, return_all, comment_characters, shutoff_dict_update):
    assert os.path.exists(file), f"File path {file} does not exist. Exiting..."
    
    if write_path is not None: #If user does not pass in write_path will attempt to create a write directory: GrievousAlignedFiles
        assert os.path.isdir(write_path), f"--write_path {write_path} does not exist as a directory. Please retry with a valid write_path. Exiting..."
    
    #load the file and instantiate proper grievous genomic object
    if file_type not in {"pvar", "ssf"}:
        logger.warning("Provided file type not in {'pvar', 'ssf'}. Attempting automatic detection of file type...")
        
        extension = file.split(".")[-1].lower()
        if extension not in {"pvar", "ssf"}:
            logger.critical("Automatic detection of file type failed. Extension does not end with pvar or ssf.")
            raise Exception("Unable to resolve file type. Please provide --file_type. Options include pvar and ssf. Exiting...")
        
        else:
            logger.warning(f"Automatic detection of file type resolved to {extension}\n")
            file_type = extension    

    if file_type == "pvar":
        gFile = Pvar(filePath = file, mappingPath = mapping, commentCharsToSkip = comment_characters)
    
    else:
        gFile = SSF(filePath = file, mappingPath = mapping, commentCharsToSkip = comment_characters)

    #Identify/Make necessary grievous database/dictionary directory and load chromosome-level database/dictionary
    grievousDictionaryPath = _Check_Dictionary(dictionary)
    dictionaryExtension = ""

    try:
        chromosomeDictionary = pd.read_parquet(os.path.join(grievousDictionaryPath, f"CHR_{gFile.Chrom}.parquet"))
        chromosomeDictionary["CHR"] = chromosomeDictionary["CHR"].astype(str)

        dictionaryExtension = "parquet"

    except:
        logger.info("parquet chromosome database not found... Trying tsv")
        assert f"CHR_{gFile.Chrom}.tsv" in os.listdir(grievousDictionaryPath), f"No valid chromosomal database file found in grievous database {dictionary} for chromosome {gFile.Chrom}. Exiting..."
        chromosomeDictionary = pd.read_csv(os.path.join(grievousDictionaryPath, f"CHR_{gFile.Chrom}.tsv"), sep = "\t", index_col = 0, dtype={'CHR': str})

        dictionaryExtension = "tsv"


    #if user fails to provide write_path attempt to make a write directory for them in the folder where the file to be aligned is:
    if write_path is None:
        fileParentDir = os.path.dirname(file)
        write_path = os.path.join(fileParentDir, "GrievousAlignedFiles")
        logger.warning(f"No write_path provided. Writing aligned files to a new directory {write_path}")
        os.makedirs(write_path, exist_ok = True) 

    else:
        if (os.path.exists(os.path.join(write_path, "GRIEVOUS_Formatted", f"FormattedAndAlignedPvar_CHR{gFile.Chrom}.tsv"))) or (os.path.exists(os.path.join(write_path, "GRIEVOUS_Formatted", f"FormattedAndAlignedSSF_CHR{gFile.Chrom}.tsv"))):
            logger.warning(f"Previous GRIEVOUS_Formatted file detected in write_path {write_path}. If this directory was used for a previous realignment, current files will be overwritten.") 

    # Make reports, reorientation, and greivousFormatted folders for output files
    os.makedirs(os.path.join(write_path, "Reports"), exist_ok = True)
    os.makedirs(os.path.join(write_path, "GRIEVOUS_Formatted"), exist_ok = True)

    if file_type == "pvar":       
        os.makedirs(os.path.join(write_path, "Reorientation"), exist_ok = True)



    #Implement Realignment

    #Employ Clean:
    gFile.Clean()

    #Employ Orient:
    gFile.Orient(chromosomeDictionary)

    #Employ Align:
    gFile.Align(write_path)

    if not shutoff_dict_update: 
        #Update grievous database/dictionary     
        gFile.UpdateDict(chromosomeDictionary,  grievousDictionaryPath, dictionaryExtension)

        #Update .updated.record file
        with open(os.path.join(grievousDictionaryPath, f"Records/CHR_{gFile.Chrom}.updated.record.txt"), "a") as updateRecord:
            updateRecord.write(file + "\n")
        

    #Write realigned files:
    gFile.Write(write_path, return_all)

    #update .aligned.record file for the given grievous database
    with open(os.path.join(grievousDictionaryPath, f"Records/CHR_{gFile.Chrom}.aligned.record.txt"), "a") as alignedRecord:
        alignedRecord.write(file + "\n")
    
    return None 


# @author: James V. Talwar

import pandas as pd
import logging
from collections import defaultdict
import os

logger = logging.getLogger(__name__)


'''
Converts a given file type to the GRIEVOUS Standard. Mapping file should at least contain all columns which deviate from standard.

Input(s):
   file: An unformatted GRIEVOUS column deviating pandas dataframe
   mappingFilePath: String corresponding to a mapping path 
Output(s):
    A formatted GRIEVOUS dataframe
'''
def FormatToStandard(file, mappingFilePath):
    mapping = pd.read_csv(mappingFilePath, delim_whitespace = True, header = None)
    mapping = dict(zip(mapping[0], mapping[1]))

    return file.rename(columns = mapping)

'''
About: This function creates a defacto graph where each CHR:POS can be considered a node and each ID an edge. 
       Non-singly connected nodes (i.e., nodes for which can reach another CHR:POS) are invalid and removed through 
       this preprocessing.

Input(s): Formatted pandas dataframe of a genomic object file (i.e., pvar or SSF)
Output(s): A processed dataframe of the genomic file filtered out for invalid SNPs that have multiple (RS)ID mappings
'''
def ExtractValidSNPs(formattedFile): 
    chrLocToID = defaultdict(set)
    idToChrLoc = defaultdict(set)
    
    for i in formattedFile.index: 
        chrLocToID[formattedFile.loc[i, "EasyCHRLOC"]].add(formattedFile.loc[i, "ID"]) 
        idToChrLoc[formattedFile.loc[i, "ID"]].add(formattedFile.loc[i, "EasyCHRLOC"]) 
        
    graph = defaultdict(lambda: defaultdict(set)) 
    
    for node,edges in chrLocToID.items():
        #Iterate through and populate the graph
        for edge in edges:
            graph[node][edge] = idToChrLoc[edge]
    
 
    keepForDownstreamProcessing = set()
    for node,edges in graph.items():
        validSNP = True
        for edge,otherNodes in edges.items():
            if not((len(otherNodes) == 1) and (next(iter(otherNodes)) == node)): #if there is ever a single instance of an ID pointing to more than one location then it is invalid and needs to be cut everywhere
                validSNP = False
        if validSNP:
            keepForDownstreamProcessing.add(node)
        
    toReturn = formattedFile[formattedFile.EasyCHRLOC.isin(keepForDownstreamProcessing)] #valid SNPs
    logger.info("{}/{} valid SNPs (with IDs not pointing to multiple locations) were extracted in preprocessing.\n".format(toReturn.shape[0], formattedFile.shape[0]))
        
    return toReturn

'''
Input(s): A dataframe of valid SNPs (i.e., SNPs where (RS)IDs are unique to a particular CHR LOC and are all valid SNPs in terms of REF/ALT) 
Output(s): A list of indexes of true biallelic SNPs 
'''
def ExtractBiallelicCandidates(chromPosDouble):
    logger.info("Beginning ExtractBiallelicCandidates() Filtering: {} SNPs under investigation... \n".format(chromPosDouble.shape[0]))
    
    #Step 0: Get sorted ref and alt allele --> This will be used in downstream filtering across file partitions
    chromPosDouble["SortedREFALT"] = ["".join(sorted(el)) for el in (chromPosDouble["REF"] + chromPosDouble["ALT"])]
    
    #Step 1: Are CHR, POS, REF, ALT jointly duplicated?
    amIDuplicatedByEverythingButID = chromPosDouble.duplicated(subset = ["CHR", "POS", "REF", "ALT"], keep = False)
    allDoubles = chromPosDouble[amIDuplicatedByEverythingButID]
    logger.info("Filter IDENTIFIED {} biallelic SNPs that were duplicated by CHR, POS, REF, ALT.".format(allDoubles.shape[0]))
    multiallelicSNPs = chromPosDouble[~amIDuplicatedByEverythingButID] #not really multiallelic SNPS, but not duplicated by chr,pos,ref,alt
    
    #Step 2: For those duplicated, everything matches but ID so drop duplicates:
    singlesTennis = allDoubles.drop_duplicates(subset = ["CHR", "POS", "REF", "ALT"]) #This can still have invalid SNPs such as double duplications
    
    #Step 3: Concatenate those SNPs that weren't jointly duplicated at all positions excepting ID with singlesTennis (i.e., those non-duplicated)
    stacked = pd.concat([singlesTennis, multiallelicSNPs], axis = 0)
    
    #Step 4: Partition into those duplicated by CHR and POS. Those unduplicated here are valid biallelic SNPs. 
    # Those duplicated need to continue to downstream filtering to identify any potential valid SNPs
    chrPosDuplication = stacked.duplicated(subset = ["CHR", "POS"], keep = False)
    noDups = stacked[~chrPosDuplication] #No duplicates by CHR, POS --> biallelic SNPs
    logger.info("Filter EXTRACTED {} biallelic SNPs that were unduplicated or were duplicated by CHR, POS, REF, ALT and were not duplicated through alternative means.".format(noDups.shape[0]))
    forFurtherFiltering = stacked[chrPosDuplication] #These are alleles that are duplicated by CHR, POS and need to proceed to step 5 
    
    del chrPosDuplication

    #Step 5: For those that were duplicated (dups) have to ask the question of are you unique by CHR, POS, SortedREFALT. If so then these are SNPs that were duplicated N number of times, but only with REF and ALT switched
    filteringDictionary = defaultdict(set) #Instantiate a checking dictionary
    for i in forFurtherFiltering.index:
        filteringDictionary[forFurtherFiltering.loc[i, "EasyCHRLOC"]].add(forFurtherFiltering.loc[i, "SortedREFALT"])
    
    realBiallelicChrLoc = set() 
    validIDs = {"AC", "AG", "AT", "CG", "CT", "GT"} #don't want any alleles that are of the form CC GG TT AA
    for k,v in filteringDictionary.items():
        if (len(v) == 1) and (next(iter(v)) in validIDs): #there is only one SortedREFALT orientation and its not of the doubled allele (e.g. CC) form
            realBiallelicChrLoc.add(k)
    
    #Get the remaining biallelic SNPs:
    sneakyBiallelics = forFurtherFiltering.EasyCHRLOC.isin(realBiallelicChrLoc)
    validSnpsAsWell = forFurtherFiltering[sneakyBiallelics] 
    logger.info("Filter IDENTIFIED AND EXTRACTED {} valid biallelic SNPs that were duplicated by CHR, POS due to duplication through REF/ALT switches.".format(validSnpsAsWell.shape[0]))

    invalidSnps = forFurtherFiltering[~sneakyBiallelics]  
    logger.info("Filter IDENTIFIED {} invalid SNPs that were duplicated by CHR, POS but had multiple alleles per position.".format(invalidSnps.shape[0]))
    
    #Now drop the duplicates from valid SNPs by those columns where duplication is relevant - namely CHR, POS, and sortedREFALT
    validSnpsAsWell = validSnpsAsWell.drop_duplicates(subset = ["CHR", "POS", "SortedREFALT"]) 
    
    
    #Step 6 (Nearly There, Nearly There...): Concatenate all the valid SNPs and return a list of indexes:
    toReturn = pd.concat([noDups, validSnpsAsWell], axis = 0)
     
    return list(toReturn.index)

'''
Input(s): A cleaned biallelic dataframe of a genomic file
Output(s): A dictionary mapping CHR:POS:REF:ALT to a given genomic file index
'''
def GenerateForwardIndices(genomicFile):
    goInThisOrder = ["CHR", "POS", "REF", "ALT"]
    grievousStandard = genomicFile[goInThisOrder].apply(lambda x: ":".join(map(str, x)), axis = 1)
    mapping = dict(zip(grievousStandard.values, grievousStandard.index))

    return mapping


'''
About: For all the keys that didn't map to grievous database/dictionary generate the alternative order - CHR:POS:ALT:REF

Input(s): temporalPincer: A dictionary of keys/indexes that were not found in grievous database (we're going both forward and backwards!)
Output(s): reversedGrievousKeys: A dictionary of CHR:POS:ALT:REF to the original input value (for which the key was CHR:POS:REF:ALT)
'''
def GenerateReverseIndices(temporalPincer):
    reversedGrievousKeys = defaultdict(str) 
    for k,v in temporalPincer.items():
        individualElements = k.split(":")
        if len(individualElements) != 4:
            logger.error("INDEX ERROR: Current index is length {} instead of 4. Exiting...".format(len(individualElements)))
            raise Exception("INDEX ERROR: Current index is length {} instead of 4. Exiting...".format(len(individualElements)))
            
        chrPosAltRef = ":".join([individualElements[0], individualElements[1], individualElements[3], individualElements[2]]) 
        reversedGrievousKeys[chrPosAltRef] = v #assign the reverse 
        
    return reversedGrievousKeys

'''
About: For all SNPs that are not indexed in a grievous chromosome db/dictionary in either orientation, ensure that the CHR:POS is not in said db/dictionary. If it is then the genomic SNP is invalid 
       as it points to different alleles than those of the db/dictionary. If not return a dictionary of CHR:LOC:REF:ALT: {CHR:STR POS:### REF:STR ALT:STR ID:STR} in the original genomic file 
       for that SNP.

Input(s):  1) snpsNotInDict: The biallelic genomic file SNPs not in the grievous db/dictionary (dataframe) 
           2) chromosomeDictionary: grievous chromosome-level db/dictionary (relating to the specific chromosome under investigation)
Output(s): 1) addendum: A dictionary of CHR:POS:REF:ALT --> {CHR:STR, POS:###, REF:STR, ALT:STR, ID:STR} for SNPs not in grievous db/dictionary 
           2) invalidSNPs: A set of SNP indexes where CHR POS is in the dictionary, but the alleles are different.
'''
def SanityCheckSNPsNotInDictionary(snpsNotInDict, chromosomeDictionary):
    goInThisOrder = ["CHR", "POS", "REF", "ALT"]
       
    uhOhSpudodios = snpsNotInDict[(snpsNotInDict.POS.isin(chromosomeDictionary.POS))] #Identified biallelic SNPs that are in the db/dictionary but with different SNPs at REF and ALT; chromosome-level so only need to filter by pos since all CHR match
    invalidSNPs = set(uhOhSpudodios.index) 
    validSNPsToAddToDict = snpsNotInDict[~snpsNotInDict.index.isin(uhOhSpudodios.index)]
    
    validSNPsToAddToDict.index = validSNPsToAddToDict[goInThisOrder].astype(str).apply(':'.join, axis=1)
    addendum = validSNPsToAddToDict[goInThisOrder + ["ID"]].T.to_dict() #subset down to only the 5 dictionary reqs  
    del validSNPsToAddToDict
    
    dictionaryCollisions = chromosomeDictionary[(chromosomeDictionary.POS.isin(uhOhSpudodios.POS))] 
    
    #Report in logs to user all SNPs that are identified as biallelic, but diverge from grievous database and thus are excluded
    for i in uhOhSpudodios.index:
        amIHere = dictionaryCollisions[dictionaryCollisions.POS == uhOhSpudodios.loc[i, "POS"]] 
        logger.info("SNP index {} (ID: {}) disagrees between pvar/ssf and database and will be removed. GRIEVOUS Database: {}; pvar/ssf: {}".format(i, uhOhSpudodios.loc[i, "ID"], list(amIHere.loc[:, "REF"] + amIHere.loc[:, "ALT"])[0], uhOhSpudodios.loc[i, "REF"] + uhOhSpudodios.loc[i, "ALT"]))
            
    del dictionaryCollisions
    
    return addendum, invalidSNPs

'''
About: Validates updated GRIEVOUS database/dictionary write contains all the biallelic SNPs not in the previous chromosome 
       alignment database/dictionary. 

Input(s):  1) updatedDictPath: String corresponding to the full path of the updated dictionary. 
           2) dictionaryAddendum: Dictionary corresponding to the biallelic SNPs (keys) and relevant information 
              that should have been added to the updated dictionary. 
           3) previousDict: Dataframe corresponding to the previous version (i.e., the one used for Orient()) of the 
              chromosome alignment dictionary.
           4) fileExtension: String in {parquet, tsv} corresponding to dictionary storage. Default parquet.
Output(s): validated: Boolean corresponding to the validation of the update dict 
'''
def ValidateDictUpdate(updatedDictPath, dictionaryAddendeum, previousDict, fileExtension = "parquet"): 
    assert fileExtension in {"parquet", "tsv"}, "Invalid database/dictionary fileExtension. Valid options are parquet or tsv."
    
    if fileExtension == "parquet":
        updatedDict = pd.read_parquet(updatedDictPath)
        updatedDict["CHR"] = updatedDict["CHR"].astype(str)
    else:
        updatedDict = pd.read_csv(updatedDictPath, sep = "\t", index_col = 0, dtype={'CHR': str})
            
    
    if len(updatedDict.index) != len(set(updatedDict.index)):
        logger.warning("Updated dictionary contains duplicated elements.")

    addOns = updatedDict[updatedDict.index.isin(dictionaryAddendeum.keys())]
    validated = addOns.shape[0] == len(dictionaryAddendeum)
    if not validated:
        logger.warning("Not all cohort identified biallelic SNPs were added to updated dictionary")
    
    validated = validated and (previousDict.shape[0] == (updatedDict.shape[0] - addOns.shape[0])) and (previousDict.shape[1] == updatedDict.shape[1])

    if not validated:
        logger.warning("Dimension inconsistency between previous and updated dictionaries.")

    return validated


'''
About: CleanReorientIndex is a method for handling duplications in pvar reorientation files. Specifically, it takes in a reorientIndex dataframe 
as extracted from Pvar.Write() and removes all duplicates (ensuring plink reorientation can be performed in cases of variant/index duplication).
The final output indexes can also be used to remove duplicates from the reorientRefAlleleThisWay dataframes. 
SSF file subset can also be used (as in SSF.Write()) to identify duplicates and generate Duplication_Reports.

Input(s):  1) duplicatedReorientIndex: reorientIndex dataframe (numerically indexed) unfiltered for duplicates.
           2) biallelicSNPs: list of identified chromosomal biallelic SNPs 
Output(s): 1) finalValidDF: A duplicate-free (i.e., filtered) reorientIndex dataframe
           2) duplicatedIDReport: dictionary corresponding to biallelic and non-biallelic IDs and their mapping to grievous index
           3) duplicatedIDCounts: dictionary corresponding to original file IDs and the number of times an ID occurs (if duplicated)
'''
def CleanReorientIndex(duplicatedReorientIndex, biallelicSnps):
    #Identify all duplicate IDs in the original file standard
    whereAreDuplications = duplicatedReorientIndex.duplicated(subset = ["ID"], keep = False)
    noDuplicates = duplicatedReorientIndex[~whereAreDuplications]
    checkForIssues = duplicatedReorientIndex[whereAreDuplications] 

    fastLookup = set(biallelicSnps) 

    #Generate a duplicated ID reports for the user:
    duplicatedIDReport = defaultdict(lambda: defaultdict(set))
    for i in checkForIssues.index:
        grievousIndex = checkForIssues.loc[i, "ProperlyFormattedName"]
        originalID = checkForIssues.loc[i, "ID"]
        if (grievousIndex in fastLookup):
            duplicatedIDReport["Biallelic_Duplicate"][originalID].add(grievousIndex)
        else:
            duplicatedIDReport["NOT_Biallelic_Duplicate"][originalID].add(grievousIndex)

    duplicatedCounts = checkForIssues.ID.value_counts()
    duplicatedIDCounts = dict(zip(duplicatedCounts.index, duplicatedCounts.values))
     
    noCompleteDuplicates = checkForIssues.drop_duplicates() #Drop duplicates if both original ID and grievous index are duplicated
    
    '''
     Now need to check if there are any duplicates by the original ID (i.e., first column) again - if so then there are biallelic SNPs doubled on the SNP chip and only 
     one is oriented the correct way as done during valid biallelic extraction to remove redundancies. For this to happen the original orientation would also need to 
     be opposite the dictionary orientation (if non-rsid; rsid duplication does not necessitate this).
    '''

    anyMoreDuplicates = noCompleteDuplicates.duplicated(subset = ["ID"], keep = False)
    canStackThese = noCompleteDuplicates[~anyMoreDuplicates]
    downstreamExtraction = noCompleteDuplicates[anyMoreDuplicates]
    
    canAlsoStackThese = pd.DataFrame()

    if downstreamExtraction.shape[0] != 0:
        logger.info(f"Duplication of original file ID, but unique grievous index case triggered. Identified {downstreamExtraction.shape[0]} instances\n")

        grievousBiallelicNumericalIndex = set()
        idHasBeenAdded = set()
    
        for i in downstreamExtraction.index: 
            grievousIndex = downstreamExtraction.loc[i, "ProperlyFormattedName"]
            if (grievousIndex in fastLookup):
                #elements are duplicated by ID, but for each duplicated ID each grievous index must be unique as have dropped complete duplicates
                grievousBiallelicNumericalIndex.add(i)

                # For instances of IDs duplicated one of which is marked as grievous biallelic (i.e., "ProperlyFormattedName" in biallelicSNPs) 
                # it is possible that the alternate orientation exists with the same ID. Need to identify those so they are not concatenated 
                # and thus duplications are included in NoDuplicates Files
                idHasBeenAdded.add(downstreamExtraction.loc[i, "ID"]) 
                

        biallelicCornerCase = downstreamExtraction[downstreamExtraction.index.isin(grievousBiallelicNumericalIndex)]

        # Handle remaining non-biallelic SNPs in downstream extraction without omission
        remainingDuplicates = downstreamExtraction[(~downstreamExtraction.index.isin(grievousBiallelicNumericalIndex)) & (~downstreamExtraction.ID.isin(idHasBeenAdded))].drop_duplicates(subset = ["ID"], keep = "first")

        canAlsoStackThese = pd.concat([biallelicCornerCase, remainingDuplicates], axis = 0)


    #stack, sort and return filtered reorient index
    finalValidDF = pd.concat([pd.concat([noDuplicates, canStackThese], axis = 0), canAlsoStackThese], axis = 0)
    finalValidDF = finalValidDF.sort_index()
    
    return finalValidDF, duplicatedIDReport, duplicatedIDCounts


# @author: James V. Talwar

import pandas as pd
import os
import logging
from ._utils import (FormatToStandard, ExtractValidSNPs, ExtractBiallelicCandidates, GenerateForwardIndices, 
                     GenerateReverseIndices, SanityCheckSNPsNotInDictionary, ValidateDictUpdate, CleanReorientIndex)

logger = logging.getLogger(__name__)

class Pvar:
    '''
    Parameters:
       filePath: String corresponding to where the file can be found (e.g., pvar)
       mappingPath: String corresponding to where column conversion standard can be found (default = None)
       commentCharsToSkip: List of strings corresponding to contiguous leading comments as are found across genomic files (default = ["##"]). 
                           
    '''
    def __init__(self, filePath, mappingPath = None, commentCharsToSkip = ["##"], *args, **kwargs):  
        #Handle when the top of the file is commented
        numLinesToSkip = 0
        with open(filePath, "r") as f:
            for line in f:
                isNotCommented = True
                for commentChar in commentCharsToSkip:
                    if line[:len(commentChar)] == commentChar:
                        numLinesToSkip += 1
                        isNotCommented = False
                        break
                
                if isNotCommented:
                    break

        if mappingPath:
            file = pd.read_csv(filePath, sep = "\t", skiprows = numLinesToSkip)
            self.file = FormatToStandard(file, mappingPath)
        else:
            self.file = pd.read_csv(filePath, sep = "\t", skiprows = numLinesToSkip)
        
        #Ensure the all key columns exist and are properly formatted to GRIEVOUS standard
        assert {"CHR", "POS", "ID", "REF", "ALT"}.intersection(set(self.file.columns)) == {"CHR", "POS", "ID", "REF", "ALT"}, "File formatting error: Mandatory columns [CHR, POS, ID, REF, ALT] are missing. Exiting..."

        self.file["CHR"] = self.file["CHR"].astype(str) #ensure consistency when handling X chromosome
        assert len(set(self.file.CHR)) == 1, "More than 1 Chromosome found in given file. Alignment expects files at the chromosome resolution. Exiting..."
        assert self.file.CHR[0] in ([str(i) for i in range(1,23)] + ["X", "Y", "MT"]), f"File CHR {self.file.CHR[0]} deviates from grievous standard. Exiting..."

        self.fileType = ".pvar" #define file type for inheritance/reusability purposes

        #Define internal class states which will be updated during the GRIEVOUS alignment process
        self._candidateBiallelicSNPIDs = None #Candidate biallelic index subset of the genomic file as updated by the Clean() method below
        self._theseIndexesNeedToFlipTheirRefAndAltAlleles = None #Indexes of SNPs that need to be reoriented - updated by Orient() below
        self._biallelicSNPIndexes = None #The GRIEVOUS identified biallelic index SNP indexes (post QC and dictionary divergence check)
        
        #Key states:  
        self.dictionaryAddendum = dict() #updated in Orient()

        #update in Align()
        self.biallelicSNPs = list()
        self.flippedSNPs = list()

    @property
    def Chrom(self):
        return self.file.CHR[0]
    
    
    '''
    About: Perform data quality filtering and generate a biallelic candidate SNP subset. 

    Input(s): None 
    Output(s): None
    Update(s): 1) self._candidateBiallelicSNPIDs: List of self.file indexes that pass genomic file only (i.e., not dictionary compared) biallelic checks.         

    '''
    def Clean(self):
        logger.info("Total number of SNPs in this Chromosome {}: {}\n".format(self.fileType, self.file.shape[0]))

        self.file["EasyCHRLOC"] = [el for el in (self.file["CHR"] + ":" + self.file["POS"].astype(str))] # helper column for ease of downstream Clean method processing
        
        #Ensure valid SNPs --> Remove all those that are not in biallelic vocabulary of {A,C,G,T}; This will also remove all variants that are not length of 1
        validSNPs = {"A", "C", "G", "T"}
        validFile = self.file[(self.file.REF.isin(validSNPs)) & (self.file.ALT.isin(validSNPs))]

        #Preprocess - remove all SNPs with IDs pointing to multiple locations
        idCleanedFile = ExtractValidSNPs(validFile)

        #Perform QC on idCleanedFile to get a set of biallelic SNP candidates (which will be compared against the grievous db/dictionary in Orient)
        biallelicCandidateIndices = ExtractBiallelicCandidates(idCleanedFile)
        
        #Check for potential AA, CC, TT, GG, alleles and remove
        ensureThingsAreValid = validFile.loc[biallelicCandidateIndices, :]
        sanityChecker = ensureThingsAreValid[ensureThingsAreValid.REF == ensureThingsAreValid.ALT].shape[0] == 0
        logger.info("Sanity Check: Are all identified biallelic candidate SNPs valid (i.e., not AA, CC, GG, TT)? {}\n".format(sanityChecker))
        
        if not sanityChecker:
            logger.info("{} AA, CC, GG, TT SNPs detected. Removing now...\n".format(ensureThingsAreValid[ensureThingsAreValid.REF == ensureThingsAreValid.ALT].shape[0]))
            ensureThingsAreValid = ensureThingsAreValid[ensureThingsAreValid.REF != ensureThingsAreValid.ALT]
    
        self._candidateBiallelicSNPIDs = list(ensureThingsAreValid.index)
        self.file = self.file.drop(["EasyCHRLOC"], axis = 1) #remove helper column

        return None

    '''
    About: Orient SNPs to the GRIEVOUS database/dictionary standard. Identify those SNPs which exist in both the db/dictionary orientation (and the reverse orientation). 
    Perform final QC on biallelic SNPs that are not in db/dictionary before adding them as biallelic SNPs and to future db/dictionary versions.
    
    Input(s): 1) chromosomeDictionary: chromosome level dataframe 
    Output(s): None
            
    Update(s):
            1) self._theseIndexesNeedToFlipTheirRefAndAltAlleles: List of given cohort genomic file indexes which exist in the opposite orientation from db/dictionary 
            2) self._biallelicSNPIndexes: List of all biallelic SNP indexes as identified by GRIEVOUS
            3) self.dictionaryAddendum: Dictionary of SNPs to be added to future GRIEVOUS db/dictionary versions

    '''
    def Orient(self, chromosomeDictionary):
        biallelicCandidateSubset = self.file.loc[self._candidateBiallelicSNPIDs, :] #true bialleleic SNPs
        
        #convert the dictionary indexes to a set for faster lookup
        doIExist = set(chromosomeDictionary.index)
        grievousOrientationToIndex = GenerateForwardIndices(biallelicCandidateSubset) 

        #Ask the question of which elements are not found in the grievous db/dictionary
        problemChildren = set([k for k in grievousOrientationToIndex]).difference(doIExist) #all the CHR:LOC:REF:ALT positions that don't match the grievous db/dictionary 
        logger.info("{}/{} SNPs were found in grievous chromosome database in their given orientation.".format(len(grievousOrientationToIndex) - len(problemChildren), biallelicCandidateSubset.shape[0]))

        #Generate keys in the other orientation
        twistIt = GenerateReverseIndices({k:grievousOrientationToIndex[k] for k in problemChildren})
        
        #Find which keys are in the dictionary but in the other orientation (and thus need to be flipped in the pvar file)
        reverseOrientationInDict = set([k for k in twistIt]).intersection(doIExist)
        self._theseIndexesNeedToFlipTheirRefAndAltAlleles = [twistIt[k] for k in reverseOrientationInDict] #These are the indexes of SNPs that need to be changed 
        logger.info("{}/{} SNPs were found in grievous chromosome database in the opposite orientation.".format(len(self._theseIndexesNeedToFlipTheirRefAndAltAlleles), biallelicCandidateSubset.shape[0]))
        
        #Get the remaining SNPs (i.e., those that are not found in mega dictionary in either orientation)
        newChallengersApproach = set([k for k in twistIt]).difference(doIExist)
        indexOfSnpsThatNeedToBeSanityCheckedBeforeAddingToMegaDictionary = [twistIt[k] for k in newChallengersApproach]
        logger.info("{}/{} SNPs were not found in grievous chromosome database. These SNPs are proceeding to db/dictionary addition validation.\n".format(len(indexOfSnpsThatNeedToBeSanityCheckedBeforeAddingToMegaDictionary), biallelicCandidateSubset.shape[0]))

        dictionaryAddendum, snpIndexesThatDisagreeWithDictionary = SanityCheckSNPsNotInDictionary(biallelicCandidateSubset.loc[indexOfSnpsThatNeedToBeSanityCheckedBeforeAddingToMegaDictionary, :], chromosomeDictionary) 
        logger.info("There were {} genomic biallelic SNPs that deviated from grievous chromosome database. Removing these...\n".format(len(snpIndexesThatDisagreeWithDictionary)))

        #Get the list of valid SNP indexes/remove invalid SNPs (those that were in mega dict, but differed by allele)
        self._biallelicSNPIndexes = [i for i in biallelicCandidateSubset.index if i not in snpIndexesThatDisagreeWithDictionary]
        
        #Update internal dictionary state
        self.dictionaryAddendum = dictionaryAddendum
          
        return None

    '''
    About: GRIEVOUS database/dictionary ALIGNs (true biallelic SNPs incorrectly oriented are swapped) the full genomic file (self.file) indexed by CHR:POS:REF:ALT (after correction). 
    It also extracts and updates the true biallelic indexes by CHR:LOC:REF:ALT (converting from the numerical indexes). 
     - NOTA BENE: This method should be called after Orient() which updates the necessary internal states self._theseIndexesNeedToFlipTheirRefAndAltAlleles and self._biallelicSNPIndexes.

    Input(s):  1) writePath: String corresponding to the directory for which to write duplication warning files 
    Output(s): None

    Update(s): 1) self.file: Aligned/Corrected and properly indexed file
               2) self.biallelicSNPs: A list of CHR:LOC:REF:ALT biallelic SNP indexes used for merging to full cohort biallelic SNP

    Write(s):  1) Warning file if any true biallelic SNPs exist as duplicated in the underlying genomic file 
'''
    def Align(self, writePath):
        assert self._biallelicSNPIndexes is not None, ".Orient() must be called before .Align()"

        #Swap alleles that are backwards and orient the Beta coefficient correctly if differs from db/dictionary
        chromosomeNumber = self.file.CHR[0] 

        swapCondition = self.file.index.isin(self._theseIndexesNeedToFlipTheirRefAndAltAlleles)
        self.file.loc[swapCondition, ["REF", "ALT"]] = self.file.loc[swapCondition, ["ALT", "REF"]].values

        if self.fileType == ".ssf": #SSF files (from which this class inherits) need to also reorient their Betas when in the opposite orientation
            self.file.loc[swapCondition, "BETA"] *= -1

        #Index the dataframe according to CHR:POS:REF:ALT now that it has been db/dictionary oriented
        goInThisOrder = ["CHR", "POS", "REF", "ALT"]
        self.file.index = [":".join([str(self.file.loc[i,col]) for col in goInThisOrder]) for i in self.file.index]
        
        #Now that indexed, oriented, and aligned get the biallelic CHR:POS:REF:ALT indexes
        formattedBiallelicIndexes = list(map(list(self.file.index).__getitem__, self._biallelicSNPIndexes))
        flippedSNPs = list(map(list(self.file.index).__getitem__, self._theseIndexesNeedToFlipTheirRefAndAltAlleles))

        #Sanity check: This can happen when SNPs are completely duplicated in the original pvar by CHR POS REF ALT; Log/Report warning SNPs to user
        if self.file[self.file.index.duplicated(keep = False)].shape[0] != 0: 
            whoseCausingProblems = self.file[self.file.index.duplicated(keep = False)]
            logger.warning("WARNING: {} SNP duplication events exist in the original file.".format(whoseCausingProblems.shape[0])) 
            biallelicProblems = set(whoseCausingProblems.index).intersection(set(formattedBiallelicIndexes))
            logger.warning("Of the SNPs duplicated by GRIEVOUS formatted index, {} of them is/are biallelic SNPs.\n".format(len(biallelicProblems)))
            if len(biallelicProblems) > 0:
                logger.warning("These biallelic SNPs are {}. Writing to WARNING_CHR{}_BiallelicDuplicates file now.\n".format(biallelicProblems, chromosomeNumber))
                pd.DataFrame(biallelicProblems).to_csv(os.path.join(writePath, "Reports/WARNING_CHR{}_BiallelicDuplicates.tsv".format(chromosomeNumber)), sep = "\t", index = False, header = None)
            #logger.warning("All duplicated index elements are:\n{}".format(whoseCausingProblems.index))
            

        #Update object biallelic SNPs and flipped SNPs
        self.biallelicSNPs = formattedBiallelicIndexes
        self.flippedSNPs = flippedSNPs

        return None 
        
    '''
    About: Updates a GRIEVOUS chromosome alignment database/dictionary after cohort orientation by writing the updated
    database/dictionary, validating the update, and then replacing the previous chromosome database/dictionary with the updated one.  

    Input(s): 1) chromosomeDictionary: Dataframe corresponding to the chromosome dictionary used for Orient().
              2) dictPath: String corresponding to the directory where the chromosome dictionary is kept.
              3) fileExtension: String in {parquet, tsv} corresponding to dictionary storage. Default parquet.
    Output(s): None
    Write(s): 
              1) An updated chromosome alignment dictionary with all biallelic SNPs for the given chromosome (identified in a 
                 given cohort, and thus ensuring orientation consistency for all subsequent cohorts that have these SNPs as well).
    '''
    def UpdateDict(self, chromosomeDictionary, dictPath, fileExtension = "parquet"):
        assert self._biallelicSNPIndexes is not None, ".Orient() must be called before .UpdateDict()"
        assert fileExtension in {"parquet", "tsv"}, "Invalid database/dictionary fileExtension. Valid options are parquet or tsv."
        assert f"CHR_{self.Chrom}.{fileExtension}" in os.listdir(dictPath), "Chromosome alignment database/dictionary not found at dictPath."

        if len(self.dictionaryAddendum) == 0:
            logger.info(f"All biallelic SNPs in CHR {self.Chrom} are in the current alignment database/dictionary. Update is not required.")
            return None

        updateName = f"CHR_{self.Chrom}_UPDATED.{fileExtension}"
        updatedDictPath = os.path.join(dictPath, updateName)

        #update the dictionary and write:
        logger.info("Adding {} total SNPs to GRIEVOUS database".format(len(self.dictionaryAddendum)))
        updatedDict = pd.concat([chromosomeDictionary, pd.DataFrame(self.dictionaryAddendum).T], axis = 0)

        if fileExtension == "parquet":
            updatedDict.to_parquet(updatedDictPath)
        else:
            updatedDict.to_csv(updatedDictPath, sep = "\t")

        #Validate the updated dictionary and replace the previous version with the updated version
        validated = ValidateDictUpdate(updatedDictPath, self.dictionaryAddendum, chromosomeDictionary, fileExtension)

        if not validated:
            os.remove(updatedDictPath)
            raise Exception(f"ERROR: All biallelic SNPs not found in previous CHR {self.Chrom} alignment dictionary were not written successfully. Exiting...")
        
        else:
            logger.info("Dictionary update validated! Replacing previous alignment dictionary with the updated version.\n")
            os.remove(os.path.join(dictPath, f"CHR_{self.Chrom}.{fileExtension}"))
            os.rename(updatedDictPath, os.path.join(dictPath, f"CHR_{self.Chrom}.{fileExtension}"))

        return None
    

    '''
    About: Write all reports as well as GRIEVOUS Formatted and reorientation files, the latter of which can be passed to 
           PLINK2 to align the underlying plink genotype files. Reorientation files are exclusive to self.fileType == '.pvar'  
    
    Input(s): 1) writePath: String corresponding to the directory where want to write these files.
              2) returnFullFile: Boolean corresponding to whether to return the full aligned pvar file (including
              any uncorrected information such as INFO columns) or the grievous PVAR column subset: [CHR, POS, ID, REF, ALT].
    
    Output(s): None

    Write(s): 
        1) The chromosome-level aligned pvar 
        2) The chromosome-level ref-allele files  (can be passed to PLINK2 after duplicate removal)
        3) The chromosome-level original pvar ID to formatted pvar index files (can be passed to PLINK2 after duplicate removal)
        4) A chromosome-level original ID SNP duplication report (if any duplicate IDs exist in the original file)
        5) The chromosome-level identified biallelic SNPs
        6) The chromosome-level subset of identified biallelic SNPs that were reoriented from cohort orientation to grievous dictionary standard (if these exist).
        
    '''
    def Write(self, writePath, returnFullFile):
        getChromosomeID = self.file.CHR[0] 
        
        #Write the formatted chromosome-level aligned pvar file (can be used to ensure consistency with new files generated by plink)
        if returnFullFile:
            self.file.to_csv(os.path.join(writePath, "GRIEVOUS_Formatted/FormattedAndAlignedPvar_CHR{}.tsv".format(getChromosomeID)), sep = "\t")
        else:
            self.file[["CHR", "POS", "ID", "REF", "ALT"]].to_csv(os.path.join(writePath, "GRIEVOUS_Formatted/FormattedAndAlignedPvar_CHR{}.tsv".format(getChromosomeID)), sep = "\t")

        #create and write reorientation files (including duplicates)
        logger.info("Writing PLINK2 update files... In total there are {} SNPs for chromosome {}.".format(self.file.shape[0], getChromosomeID))

        refAlleleFile = self.file[["ID", "REF"]]
        refAlleleFile.to_csv(os.path.join(writePath, "Reorientation/ReorientRefAlleleThisWay_CHR{}.tsv".format(getChromosomeID)), sep = "\t", index = False, header = None)

        logger.info("Finished writing ref allele file...")
    
        oldIDToFormattedID = self.file[["ID"]].copy()
        oldIDToFormattedID["ProperlyFormattedName"] = oldIDToFormattedID.index
        oldIDToFormattedID.to_csv(os.path.join(writePath, "Reorientation/ReorientIndex_CHR{}.tsv".format(getChromosomeID)), sep = "\t", index = False, header = None)

        logger.info("Finished writing ID update file...")

        #Generate cleaned non-duplicated ReorientRefAlleleThisWay_ and ReorientIndex_ files for Plink realignment
        oldIDToFormattedID.index = [i for i in range(len(oldIDToFormattedID.index))]
        noDuplicateReorientIndex, duplicatedIDReport, duplicatedIDCounts = CleanReorientIndex(oldIDToFormattedID, self.biallelicSNPs)

        refAlleleFile.index = [i for i in range(len(refAlleleFile.index))]
        noDuplicateRefAlleleFile = refAlleleFile[refAlleleFile.index.isin(noDuplicateReorientIndex.index)] #numerical indexing is the same across the 2 files as both subset from self.file columns

        noDuplicateReorientIndex.to_csv(os.path.join(writePath, "Reorientation/NoDuplicates_ReorientIndex_CHR{}.tsv".format(getChromosomeID)), sep = "\t", index = False, header = None)
        noDuplicateRefAlleleFile.to_csv(os.path.join(writePath, "Reorientation/NoDuplicates_ReorientRefAlleleThisWay_CHR{}.tsv".format(getChromosomeID)), sep = "\t", index = False, header = None)
        
        #Generate duplication report (if there are indeed duplicates):
        if len(duplicatedIDCounts) > 0:
            with open(os.path.join(writePath, "Reports/CHR{}_Duplication_Report.txt".format(getChromosomeID)), "a") as duplicationReport:
                duplicationReport.write("Original file duplication ID incidences:\n\n")
                
                for k,v in duplicatedIDCounts.items():
                    duplicationReport.write("\t".join([k,str(v)]) + "\n")
                
                duplicationReport.write("\nOriginal ID to grievous index unique incidence mappings:\n")

                for k,v in duplicatedIDReport.items():
                    duplicationReport.write("\n" + k + "\n\n")
                    for originalID, grievousIndexes in v.items():
                        duplicationReport.write("\t".join([originalID , str(grievousIndexes)]) + "\n")   


        logger.info("Duplicate removal for ReorientIndex and ReorientRefAlleleThisWay complete...\n")

        logger.info("Writing biallelic SNPs and subset of biallelic SNPs that were reoriented from cohort specific orientation to grievous database standard\n")
        
        pd.DataFrame(self.biallelicSNPs, columns = ["BiallelicSNPs"]).to_csv(os.path.join(writePath, "Reports/CHR{}_BiallelicSNPs.tsv".format(getChromosomeID)), sep = "\t", index = False)
        if len(self.flippedSNPs) > 0:
            pd.DataFrame(self.flippedSNPs, columns = ["FlippedSNPs"]).to_csv(os.path.join(writePath, "Reports/CHR{}_FlippedSNPs.tsv".format(getChromosomeID)), sep = "\t", index = False)

        logger.info("GRIEVOUS Alignment for CHR {} complete\n".format(getChromosomeID))

        return None